/* builtin.c   by Michael Thorpe   2006-03-12 */

#include <ctype.h>
#include <limits.h>
#include "lisp.h"

static obj *divideby0error;
static obj *exitloop;
static obj *nextloop;
static obj *overflowerror;

static obj *add(obj *a,obj *b) {
	long result;
	int overflow=0;
	obj *r;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
	result=a->value.i+b->value.i;
	if((a->value.i^result)<0 && (b->value.i^result)<0)
		overflow=1;
	decref(a);
	decref(b);
	r=newobj(&integer_objtype);
	if(r) {
		r->value.i=result;
		if(overflow)
			return(throw(incref(overflowerror),r));
	}
	return(r);
}

static obj *bitand(obj *a,obj *b) {
	long result;
	obj *r;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
	result=a->value.i&b->value.i;
	decref(a);
	decref(b);
	r=newobj(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

static obj *bitnot(obj *a) {
	long result;
	obj *r;

	if(a->type != &integer_objtype)
		return(throwtypeerror(a));
	result=~a->value.i;
	decref(a);
	r=newobj(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

static obj *bitor(obj *a,obj *b) {
	long result;
	obj *r;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
	result=a->value.i|b->value.i;
	decref(a);
	decref(b);
	r=newobj(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

static obj *bitxor(obj *a,obj *b) {
	long result;
	obj *r;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
	result=a->value.i^b->value.i;
	decref(a);
	decref(b);
	r=newobj(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

static obj *div(obj *a,obj *b) {
	long result;
	obj *r;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
	if(b->value.i==0) {
		decref(b);
		return(throw(incref(divideby0error),a));
	}
	result=a->value.i/b->value.i;
	decref(a);
	decref(b);
	r=newobj(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

static obj *eq(obj *a,obj *b) {
	decref(a);
	decref(b);
	return(incref((a==b)?&trueobj:&nilobj));
}

static obj *ifelse(scope *s,obj *cond,obj *iftrue,obj *iffalse) {
	decref(cond);
	if(cond==&nilobj) {
		decref(iftrue);
		return(evaluate(s,iffalse));
	} else {
		decref(iffalse);
		return(evaluate(s,iftrue));
	}
}

static obj *isasciip(obj *a) {
	int result;

	if(a->type != &integer_objtype)
		return(throwtypeerror(a));
	if(a->value.i<0 || a->value.i>UCHAR_MAX)
		return(throwrangecheck(a));
	result=isascii((int)a->value.i);
	decref(a);
	return(incref(result?&trueobj:&nilobj));
}

static obj *isdigitp(obj *a) {
	int result;

	if(a->type != &integer_objtype)
		return(throwtypeerror(a));
	if(a->value.i<0 || a->value.i>UCHAR_MAX)
		return(throwrangecheck(a));
	result=isdigit((int)a->value.i);
	decref(a);
	return(incref(result?&trueobj:&nilobj));
}

static obj *isspacep(obj *a) {
	int result;

	if(a->type != &integer_objtype)
		return(throwtypeerror(a));
	if(a->value.i<0 || a->value.i>UCHAR_MAX)
		return(throwrangecheck(a));
	result=isspace((int)a->value.i);
	decref(a);
	return(incref(result?&trueobj:&nilobj));
}

static obj *lambda(obj *args,obj *prog) {
	obj *o;

	o=newobj(&func_objtype);
	if(o) {
		o->value.func.args=args;
		o->value.func.prog=prog;
	} else {
		decref(args);
		decref(prog);
	}
	return(o);
}

static obj *loop(scope *s,obj *o) {
	obj *p;

	while(1) {
		p=evaluate(incscoperef(s),incref(o));
		if(p) {
			decref(p);
			continue;
		}
		if(exception != nextloop)
			break;
		decref(exception);
		exception=incref(&nilobj);
		decref(exceptionvalue);
		exceptionvalue=incref(&nilobj);
	}
	decref(o);
	decscoperef(s);
	if(exception==exitloop) {
		decref(exception);
		exception=incref(&nilobj);
		p=exceptionvalue;
		exceptionvalue=incref(&nilobj);
		return(p);
	}
	return(0);
}

static obj *map(scope *s,obj *var,obj *expr,obj *list) {
	obj *new,*old,*start,*tmp;
	unsigned long length;

	s=newscope(s);
	if(!s)
		goto fail_noscope;
	length=0;
	for(new=list;&cons_objtype==new->type;new=new->value.c.d)
		length++;
	if(!length) {
		if(real_define(s,var,list))
			goto fail_nostart;
		decref(var);
		decref(list);
		return(evaluate(s,expr));
	}
	start=newlist(length);
	if(!start)
		goto fail_nostart;
	new=start;
	old=list;
	while(1) {
		if(real_define(s,var,CAR(old)))
			goto fail;
		tmp=evaluate(incscoperef(s),incref(expr));
		if(!tmp)
			goto fail;
		decref(CAR(new));
		CAR(new)=tmp;
		old=CDR(old);
		if(&cons_objtype != old->type)
			break;
		new=CDR(new);
	}
	if(&nilobj != old) {
		if(real_define(s,var,old))
			goto fail;
		tmp=evaluate(incscoperef(s),incref(expr));
		if(!tmp)
			goto fail;
		decref(CDR(new));
		CDR(new)=tmp;
	}
	decscoperef(s);
	decref(var);
	decref(expr);
	decref(list);
	return(start);
fail:
	decref(start);
fail_nostart:
	decscoperef(s);
fail_noscope:
	decref(var);
	decref(expr);
	decref(list);
	return(0);
}

static obj *mul(obj *a,obj *b) {
	long result;
	obj *r;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
/* FIXME: Check for overflow */
	result=a->value.i*b->value.i;
	decref(a);
	decref(b);
	r=newobj(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

#define NUMCMP_LT 1
#define NUMCMP_EQ 2
#define NUMCMP_GT 4
static obj *numcmp(obj *a,obj *b,int desired) {
	int cmp;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
	cmp=NUMCMP_EQ;
	if(a->value.i < b->value.i)
		cmp=NUMCMP_LT;
	if(a->value.i > b->value.i)
		cmp=NUMCMP_GT;
	decref(a);
	decref(b);
	return(incref((cmp&desired)?&trueobj:&nilobj));
}

static obj *num_lt(obj *a,obj *b) {
	return(numcmp(a,b,NUMCMP_LT));
}

static obj *num_le(obj *a,obj *b) {
	return(numcmp(a,b,NUMCMP_LT|NUMCMP_EQ));
}

static obj *num_ne(obj *a,obj *b) {
	return(numcmp(a,b,NUMCMP_LT|NUMCMP_GT));
}

static obj *num_eq(obj *a,obj *b) {
	return(numcmp(a,b,NUMCMP_EQ));
}

static obj *num_ge(obj *a,obj *b) {
	return(numcmp(a,b,NUMCMP_EQ|NUMCMP_GT));
}

static obj *num_gt(obj *a,obj *b) {
	return(numcmp(a,b,NUMCMP_GT));
}

static obj *prog2of2(scope *s,obj *a,obj *b) {
	decref(a);
	return(evaluate(s,b));
}

static obj *sub(obj *a,obj *b) {
	long result;
	int overflow=0;
	obj *r;

	if(a->type != &integer_objtype) {
		decref(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		decref(a);
		return(throwtypeerror(b));
	}
	result=a->value.i - b->value.i;
	if((a->value.i^b->value.i)<0 && (a->value.i^result)<0)
		overflow=1;
	decref(a);
	decref(b);
	r=newobj(&integer_objtype);
	if(r) {
		r->value.i=result;
		if(overflow)
			return(throw(incref(overflowerror),r));
	}
	return(r);
}

static obj *tostring(obj *o) {
	return((o)->type->tostring(o));
}

int builtin_init() {
	if(!(divideby0error=newword("divide-by-zero",14))
	|| !(exitloop=newword("exit-loop",9))
	|| !(nextloop=newword("next-loop",9))
	|| !(overflowerror=newword("overflow",8))
	|| store_builtin_ns2("*",mul)
	|| store_builtin_ns2("+",add)
	|| store_builtin_ns2("-",sub)
	|| store_builtin_ns2("/",div)
	|| store_builtin_ns2("<",num_lt)
	|| store_builtin_ns2("<=",num_le)
	|| store_builtin_ns2("<>",num_ne)
	|| store_builtin_ns2("=",num_eq)
	|| store_builtin_ns2(">",num_gt)
	|| store_builtin_ns2(">=",num_ge)
	|| store_builtin_ns2("bit-and",bitand)
	|| store_builtin_ns1("bit-not",bitnot)
	|| store_builtin_ns2("bit-or",bitor)
	|| store_builtin_ns2("bit-xor",bitxor)
	|| store_builtin_ns2("eq?",eq)
	|| store_builtin_s3("#ifelse",ifelse)
	|| store_builtin_ns1("isascii",isasciip)
	|| store_builtin_ns1("isdigit",isdigitp)
	|| store_builtin_ns1("isspace",isspacep)
	|| store_builtin_ns2("#lambda",lambda)
	|| store_builtin_s1("#loop",loop)
	|| store_builtin_s3("#map",map)
	|| store_builtin_s2("#prog2of2",prog2of2)
	|| store_builtin_ns1("->string",tostring))
		return(1);
	INC_KNOWN_OBJECTS(4);
	return(0);
}
